package kow.functional

object Functions01 {
  def main(args: Array[String]): Unit = {

    //------------------------------------------------------------------
    //full function in the domain of int
    val inc = (a: Int) => a + 1
    println(inc(1))

    //another full function
    val add = (a: Int, b: Int) => a + b
    println(add(3, 4))

    //------------------------------------------------------------------

    //partial function
    val one: PartialFunction[Int, String] = { case 1 => "one" }
    one.isDefinedAt(1)
    one.isDefinedAt(2)

    //another exmaple of partial function

    val divide = new PartialFunction[Int, Int] {
      def apply(x: Int) = 42 / x
      def isDefinedAt(x: Int) = x != 0
    }

    divide.isDefinedAt(1)
    if (divide.isDefinedAt(1)) divide(1)
    divide.isDefinedAt(0) // should return false

    //chaining partial functions
    // converts 1 to "one", etc., up to 5
    val convert1to5 = new PartialFunction[Int, String] {
      val nums = Array("one", "two", "three", "four", "five")
      def apply(i: Int) = nums(i - 1)
      def isDefinedAt(i: Int) = i > 0 && i < 6
    }

    // converts 6 to "six", etc., up to 10
    val convert6to10 = new PartialFunction[Int, String] {
      val nums = Array("six", "seven", "eight", "nine", "ten")
      def apply(i: Int) = nums(i - 6)
      def isDefinedAt(i: Int) = i > 5 && i < 11
    }

    val handle1to10 = convert1to5 orElse convert6to10
    handle1to10(3)

    //------------------------------------------------------------------
    //Higher order functions

    def sumInts(a: Int, b: Int): Int =
      if (a > b) 0 else a + sumInts(a + 1, b)

    def cube(x: Int): Int = x * x * x

    def sumCubes(a: Int, b: Int): Int =
      if (a > b) 0 else cube(a) + sumCubes(a + 1, b)

    def sum(f: Int => Int, a: Int, b: Int): Int =
      if (a > b) 0
      else f(a) + sum(f, a + 1, b)

    def id(x: Int): Int = x
    def sumInts2(a: Int, b: Int) = sum(id, a, b)
    def sumCubes2(a: Int, b: Int) = sum(cube, a, b)

    //many uses specially in collection libraries
    val listOfNumbers = List(3, 4, 5, 0, 6)
    val nonzeroList = listOfNumbers.filter { n => n != 0 }
    val addOneList = listOfNumbers.map { n => n + 1 }

    //------------------------------------------------------------------
    //Anonymous function

    (x: Int) => x + 1

    //------------------------------------------------------------------
    // closures

    var freeVariable = 5

    val func1 = (a: Int) => freeVariable + a
    func1(4)

    freeVariable = 10
    val funct2 = (a: Int) => freeVariable + a
    funct2(4)

    //------------------------------------------------------------------
    //curried functions
    val add2 = (a: Int, b: Int) => a + b
    val carryAddition: (Int) => (Int) => Int = add2.curried
    val inc2 = carryAddition(2)
    inc2(10)
    //------------------------------------------------------------------

    import scala.annotation.tailrec

    def factorial(n: Int) = {
      @tailrec
      def loop(start: Int, product: Int): Int = {
        if (start > n) product
        else loop(start + 1, product * start)
      }
      loop(1, 1)
    }

    def factorial2(n: Int): Int = {
      @tailrec def factorialAcc(acc: Int, n: Int): Int = {
        if (n <= 1) acc
        else factorialAcc(n * acc, n - 1)
      }
      factorialAcc(1, n)
    }

    //------------------------------------------------------------------

    //------------------------------------------------------------------

  }
}
